@@ -1,67 +1,80 @@
 import os
 import csv
-from collections import OrderedDict
-
+from collections import OrderedDict, defaultdict
 import numpy as np
 import torch
 from tqdm.auto import tqdm
 
 from torch.utils.tensorboard import SummaryWriter
-
-from deepml.predict import ImageRegressionPredictor
-from deepml.predict import ImageClassificationPredictor
-from deepml.predict import SemanticSegmentationPredictor
-
+from deepml.tasks import Predictor
 from deepml import utils
 
 
 class Learner:
 
-    def __init__(self, model, optimizer, model_save_path, model_file_name='latest_model.pt',
-                 load_saved_model=False, load_optimizer_state=True, use_gpu=False, classes=None):
+    def __init__(self, predictor, optimizer, criterion, load_optimizer_state=False):
+        """
+        Training class for learning a model weights using predictor and optimizer.
 
-        if model is None:
-            raise ValueError('Model cannot be None.')
+        :param predictor: Object of sub class deepml.tasks.Predictor
+        :param optimizer: The optimizer from torch.optim
+        :param criterion: The loss function
+        :param load_optimizer_state: Weather to load optimizer state to resume model training. Default is False.
+                                     If true, optimizer state is loaded with load_state_dict and history of epoch.
+        """
 
-        if optimizer is None:
-            raise ValueError('Optimizer cannot be None.')
+        assert isinstance(predictor, Predictor)
+        assert isinstance(optimizer, torch.optim.Optimizer)
+        assert isinstance(criterion, torch.nn.Module)
 
-        self.__model = model
+        self.__predictor = predictor
+        self.__model = self.__predictor.model
+        self.__model_dir = self.__predictor.model_dir
+        self.__model_file_name = self.__predictor.model_file_name
         self.__optimizer = optimizer
-        self.model_save_path = model_save_path
+        self.__criterion = criterion
         self.epochs_completed = 0
         self.best_val_loss = np.inf
+        self.history = defaultdict(list)
 
-        os.makedirs(self.model_save_path, exist_ok=True)
-        self.writer = SummaryWriter(os.path.join(self.model_save_path,
-                                                 utils.find_new_run_dir_name(self.model_save_path)))
-        self.device = torch.device("cuda:0" if use_gpu and torch.cuda.is_available() else "cpu")
+        os.makedirs(self.__model_dir, exist_ok=True)
+        self.writer = SummaryWriter(os.path.join(self.__model_dir,
+                                                 utils.find_new_run_dir_name(self.__model_dir)))
 
-        self.__predictor = None
-        self.__classes = classes
         self.__metrics_dict = OrderedDict({'loss': 0})
 
-        if load_saved_model and model_file_name is not None:
-            self.load_saved_model(os.path.join(self.model_save_path, model_file_name),
-                                  load_optimizer_state)
+        if load_optimizer_state:
+            self.__load_optimizer_state()
 
-        self.set_optimizer(self.__optimizer)
+        self.__device = self.__predictor.device
+        self.set_device(self.__device)
 
     def set_optimizer(self, optimizer):
-        if optimizer is None:
-            raise ValueError('Optimizer cannot be None.')
+        assert isinstance(optimizer, torch.optim.Optimizer)
+        self.__optimizer = optimizer
 
-        self.__model = self.__model.to(self.device)
-        self.__optimizer = optimizer.__class__(self.__model.parameters(),
-                                               **optimizer.defaults)
+    def set_criterion(self, criterion):
+        assert isinstance(criterion, torch.nn.Module)
+        self.__criterion = criterion
 
-    def load_saved_model(self, model_path, load_optimizer_state=False):
+    def set_device(self, device):
+        assert isinstance(device, str) and device.lower() in ['cpu', 'cuda']
+        device = device.lower()
+        self.__device = device if device == "cuda" and torch.cuda.is_available() else "cpu"
+
+        self.__model.to(self.__device)
+
+        for optim_state_values_dict in self.__optimizer.state.values():
+            for key in optim_state_values_dict:
+                if type(optim_state_values_dict[key]) == torch.Tensor:
+                    optim_state_values_dict[key] = optim_state_values_dict[key].to(self.__device)
+
+    def __load_optimizer_state(self):
+        model_path = os.path.join(self.__model_dir, self.__model_file_name)
         if os.path.exists(model_path):
-            print('Loading Saved Model Weights.')
-            state_dict = torch.load(model_path)
-            self.__model.load_state_dict(state_dict['model'])
-
-            if load_optimizer_state and 'optimizer' in state_dict:
+            state_dict = (torch.load(model_path) if torch.cuda.is_available()
+                          else torch.load(model_path, map_location=torch.device('cpu')))
+            if 'optimizer' in state_dict:
                 if state_dict['optimizer'] == self.__optimizer.__class__.__name__:
                     self.__optimizer.load_state_dict(state_dict['optimizer_state'])
                 else:
@@ -73,14 +86,6 @@ class Learner:
 
             if 'metrics' in state_dict and 'val_loss' in state_dict['metrics']:
                 self.best_val_loss = state_dict['metrics']['val_loss']
-
-            if 'classes' in state_dict:
-                self.__classes = state_dict['classes']
-
-            if 'predictor' in state_dict:
-                # instantiate predictor class
-                self.__predictor = state_dict['predictor'](model=self.__model, classes=self.__classes)
-
         else:
             print(f'{model_path} does not exist.')
 
@@ -99,19 +104,15 @@ class Learner:
         if type(train_loss) == float and type(val_loss) == float:
             save_dict['metrics'] = {'train_loss': train_loss, 'val_loss': val_loss}
 
-        if self.__predictor is not None:
-            save_dict['predictor'] = self.__predictor.__class__
+        save_dict['criterion'] = self.__criterion.__class__.__name__
 
-        if self.__classes is not None:
-            save_dict['classes'] = self.__classes
-
-        filepath = os.path.join(self.model_save_path, model_file_name)
+        filepath = os.path.join(self.__model_dir, model_file_name)
         torch.save(save_dict, filepath)
-        self.__model.to(self.device)
 
+        self.__model.to(self.__device)
         return filepath
 
-    def validate(self, criterion, loader, metrics=None):
+    def validate(self, loader, criterion, metrics=None):
         if loader is None:
             raise Exception('Loader cannot be None.')
 
@@ -122,13 +123,14 @@ class Learner:
         bar = tqdm(total=len(loader), desc="{:12s}".format('Validation'))
 
         with torch.no_grad():
-            for batch_index, (X, y) in enumerate(loader):
+            for batch_index, (x, y) in enumerate(loader):
 
-                X = X.to(self.device)
-                y = y.to(self.device)
-                outputs = self.__model(X)
+                outputs = self.__predictor.predict_batch(x)
 
-                if outputs.shape[1] == 1:
+                if isinstance(y, torch.Tensor):
+                    y = y.to(self.__device)
+
+                if isinstance(outputs, torch.Tensor) and outputs.ndim == 2 and outputs.shape[1] == 1:
                     y = y.view_as(outputs)
 
                 loss = criterion(outputs, y)
@@ -138,27 +140,13 @@ class Learner:
                 self.__update_metrics(outputs, y, metrics, batch_index + 1)
 
                 bar.update(1)
-                bar.set_postfix({name: f'{round(value, 2)}' for name, value in self.__metrics_dict.items()})
+                bar.set_postfix({name: f'{round(value, 4)}' for name, value in self.__metrics_dict.items()})
 
         return self.__metrics_dict
 
-    def __infer_predictor(self, x):
-
-        self.__model.eval()
-        with torch.no_grad():
-            prediction = self.__model(x.to(self.device)).cpu()
-
-            if self.__classes is None:
-                try:
-                    prediction = prediction.item()
-                    return ImageRegressionPredictor(self.__model)
-                except ValueError:
-                    pass
-
-            if prediction.ndim == 2:
-                return ImageClassificationPredictor(self.__model, classes=self.__classes)
-            else:
-                return SemanticSegmentationPredictor(self.__model, classes=self.__classes)
+    def set_predictor(self, predictor):
+        assert isinstance(predictor, Predictor)
+        self.__predictor = predictor
 
     def __init_metrics(self, metrics):
         if metrics is None:
@@ -175,10 +163,6 @@ class Learner:
         if metrics is None:
             return
 
-        # Update metrics
-        outputs = outputs.to(self.device)
-        targets = targets.to(self.device)
-
         for metric_name, metric_instance in metrics:
             self.__metrics_dict[metric_name] = self.__metrics_dict[metric_name] + \
                                                ((metric_instance(outputs, targets).item() - self.__metrics_dict[
@@ -186,66 +170,95 @@ class Learner:
 
     def __write_metrics_to_tensorboard(self, tag, global_step):
         for name, value in self.__metrics_dict.items():
-            self.writer.add_scalar(f'{tag}/{name}', value, global_step)
+            self.writer.add_scalar(f'{name}/{tag}', value, global_step)
 
-    def __write_lr_to_tensorboard(self, global_step):
-        # Write lr to tensor-board
-        param_group = self.__optimizer.param_groups[0]
-        self.writer.add_scalar('learning_rate', param_group['lr'], global_step)
+    def __write_history(self, stage):
+        for name, value in self.__metrics_dict.items():
+            self.history[f"{stage}_{name}"].append(value)
 
-    def fit(self, criterion, train_loader, val_loader=None, epochs=10, steps_per_epoch=None,
-            save_model_after_every_epoch=5, lr_scheduler=None, image_inverse_transform=None,
-            metrics=None, tboard_img_size=224):
+    def __write_lr(self, global_step):
+        # Write lr to tensor-board and history dict
+        if len(self.__optimizer.param_groups) == 1:
+            param_group = self.__optimizer.param_groups[0]
+            self.writer.add_scalar('learning_rate', param_group['lr'], global_step)
+            self.history['learning_rate'].append(param_group['lr'])
+        else:
+            for index, param_group in enumerate(self.__optimizer.param_groups):
+                self.writer.add_scalar(f'learning_rate/param_group_{index}', param_group['lr'],
+                                       global_step)
+                self.history[f'learning_rate/param_group_{index}'].append(param_group['lr'])
+
+    def __write_graph_to_tensorboard(self, loader):
+
+        if not loader:
+            return
+
+        # Write graph to tensorboard
+        temp_x = None
+        for X, _ in loader:
+            temp_x = X
+            break
+
+        temp_x = self.__predictor.models_input_to_device(temp_x)
+
+        with torch.no_grad():
+            self.__model.eval()
+            try:
+                self.writer.add_graph(self.__model, temp_x)
+            except Exception as e:
+                print("Warning: Failed to write graph to tensorboard.", e)
+
+    def fit(self, train_loader, val_loader=None, epochs=10, steps_per_epoch=None,
+            save_model_after_every_epoch=5, lr_scheduler=None, lr_scheduler_step_policy='epoch',
+            metrics=None, image_inverse_transform=None, tboard_img_size=224):
 
         """
-        Starts training the model on specified train loader
+        Trains the model on specified train loader for specified number of epochs.
 
         Parameters
         ----------
-        :param criterion: loss function to optimize
-
         :param train_loader: The torch.utils.data.DataLoader for model to train on.
 
         :param val_loader: The torch.utils.data.DataLoader for model to validate on.
-        Default is None.
+                           Default is None.
 
         :param epochs: int The number of epochs to train. Default is 10
 
-        :param steps_per_epoch: Should be around len(train_loader),
-        so that every example in the dataset gets covered in each epoch.
+        :param steps_per_epoch: Should be around len(train_loader), so that every example in the
+                                dataset gets covered in each epoch.
 
         :param save_model_after_every_epoch: To save the model after every number of completed epochs
-        Default is 5.
+                                            Default is 5.
 
         :param lr_scheduler: the learning rate scheduler, default is None.
 
-        :param image_inverse_transform: It denotes reverse transformations of image normalization so that images
+        :param lr_scheduler_step_policy: It is the time when lr_scheduler.step() would be called.
+                                         Default is "epoch" policy.
+                                         Use "batch" policy if you want lr_scheduler.step() to be
+                                         called after each gradient step.
 
-        can be displayed on tensor board. Default is deepml.transforms.ImageNetInverseTransform() which is
-        an inverse of ImageNet normalization.
+        :param image_inverse_transform: It denotes reverse transformations of image normalization so that images
+                                        can be displayed on tensor board.
+                                        Default is deepml.transforms.ImageNetInverseTransform() which is
+                                        an inverse of ImageNet normalization.
 
         :param metrics: list of tuples ('metric_name', metric instance) to monitor.
-        Metric name is used as label for logging metric value to console and tensorboard.
-        Metric instance must be subclass of torch.nn.Module, implements forward function and
-        returns calculated value.
+                        Metric name is used as label for logging metric value to console and tensorboard.
+                        Metric instance must be subclass of torch.nn.Module, implements forward function and
+                        returns calculated value.
 
         :param tboard_img_size:  image size to use for writing images to tensorboard
         """
         if steps_per_epoch is None:
             steps_per_epoch = len(train_loader)
 
-        self.__model = self.__model.to(self.device)
-        criterion = criterion.to(self.device)
+        assert steps_per_epoch <= len(train_loader), "Steps per epoch should not be greater than len(train_loader)"
 
-        if self.__predictor is None:
-            x, _ = train_loader.dataset[0]
-            # Add batch dimension
-            x = x.unsqueeze(dim=0)
-            self.__predictor = self.__infer_predictor(x)
+        self.__model.to(self.__device)
+        self.__criterion = self.__criterion.to(self.__device)
 
         # Write graph to tensorboard
-        temp_x, _ = train_loader.dataset[0]
-        self.writer.add_graph(self.__model, temp_x.unsqueeze(dim=0).to(self.device))
+        self.__write_graph_to_tensorboard(val_loader)
 
         # Check valid metrics types
         if metrics:
@@ -253,11 +266,15 @@ class Learner:
                 if not (isinstance(metric_instance, torch.nn.Module) and hasattr(metric_instance, 'forward')):
                     raise TypeError(f'{metric_instance.__class__} is not supported')
 
+        # Check valid policy for lr_scheduler
+        if lr_scheduler is not None:
+            lr_scheduler_step_policy = lr_scheduler_step_policy.lower()
+            assert isinstance(lr_scheduler_step_policy, str) and lr_scheduler_step_policy in ['epoch', 'batch']
+
         # Replace all metrics during call to learner fit
         self.__metrics_dict = OrderedDict({'loss': 0})
 
         train_loss = 0
-        epoch = 0
         epochs = self.epochs_completed + epochs
         for epoch in range(self.epochs_completed, epochs):
             print('Epoch {}/{}:'.format(epoch + 1, epochs))
@@ -266,37 +283,35 @@ class Learner:
 
             # Iterate over batches
             step = 0
-            lr_step_done = False
 
             # init all metrics with zeros
             self.__metrics_dict['loss'] = 0
             self.__init_metrics(metrics)
 
             # Write current lr to tensor-board
-            self.__write_lr_to_tensorboard(epoch + 1)
+            self.__write_lr(epoch + 1)
 
             bar = tqdm(total=steps_per_epoch, desc="{:12s}".format('Training'))
-            for batch_index, (X, y) in enumerate(train_loader):
-
-                X = X.to(self.device)
+            for batch_index, (x, y) in enumerate(train_loader):
 
                 # zero the parameter gradients
                 self.__optimizer.zero_grad()
 
-                outputs = self.__model(X)
+                outputs = self.__predictor.predict_batch(x)
 
-                if outputs.shape[1] == 1:
+                if isinstance(y, torch.Tensor):
+                    y = y.to(self.__device)
+
+                if isinstance(outputs, torch.Tensor) and outputs.ndim == 2 and outputs.shape[1] == 1:
                     y = y.view_as(outputs)
 
-                y = y.to(self.device)
-                loss = criterion(outputs, y)
+                loss = self.__criterion(outputs, y)
                 loss.backward()
 
                 self.__optimizer.step()
 
-                if isinstance(lr_scheduler, torch.optim.lr_scheduler.CyclicLR):
+                if lr_scheduler and lr_scheduler_step_policy == "batch":
                     lr_scheduler.step()
-                    lr_step_done = True
 
                 step = step + 1
                 self.__metrics_dict['loss'] = self.__metrics_dict['loss'] + ((loss.item() - self.__metrics_dict['loss'])
@@ -304,40 +319,37 @@ class Learner:
                 # Update metrics
                 self.__update_metrics(outputs, y, metrics, step)
                 bar.update(1)
-                bar.set_postfix({name: f'{round(value, 2)}' for name, value in self.__metrics_dict.items()})
+                bar.set_postfix({name: f'{round(value, 4)}' for name, value in self.__metrics_dict.items()})
+
+                if (batch_index + 1) % steps_per_epoch == 0:
+                    break
 
             self.epochs_completed = self.epochs_completed + 1
 
             # Write some sample training images to tensorboard
-            X, y = utils.get_random_samples_batch_from_loader(train_loader)
-            X, y = X.to(self.device), y.to(self.device)
-            self.__predictor.write_prediction_to_tensorboard('Train', (X, y),
+            self.__predictor.write_prediction_to_tensorboard('train', train_loader,
                                                              self.writer, image_inverse_transform,
                                                              self.epochs_completed, img_size=tboard_img_size)
 
             train_loss = self.__metrics_dict['loss']
-            self.__write_metrics_to_tensorboard('Train', self.epochs_completed)
-
+            self.__write_metrics_to_tensorboard('train', self.epochs_completed)
+            self.__write_history('train')
+            message = f"Training Loss: {train_loss:.4f} "
             val_loss = np.inf
             if val_loader is not None:
-                self.validate(criterion, val_loader, metrics)
+                self.validate(val_loader, self.__criterion, metrics)
                 val_loss = self.__metrics_dict['loss']
-                self.__write_metrics_to_tensorboard('Val', self.epochs_completed)
-
+                self.__write_metrics_to_tensorboard('val', self.epochs_completed)
+                self.__write_history('val')
+                message = message + f"Validation Loss: {val_loss:.4f} "
                 # write random val images to tensorboard
-                X, y = utils.get_random_samples_batch_from_loader(val_loader)
-                X, y = X.to(self.device), y.to(self.device)
-                self.__predictor.write_prediction_to_tensorboard('Val', (X, y),
+                self.__predictor.write_prediction_to_tensorboard('val', val_loader,
                                                                  self.writer, image_inverse_transform,
                                                                  self.epochs_completed,
                                                                  img_size=tboard_img_size)
-
-                if isinstance(lr_scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
-                    lr_scheduler.step(val_loss)
-                    lr_step_done = True
-
                 # Save best validation model
                 if val_loss < self.best_val_loss:
+                    message = message + "[Saving best validation model]"
                     self.best_val_loss = val_loss
                     self.save('best_val_model.pt',
                               save_optimizer_state=True,
@@ -345,12 +357,16 @@ class Learner:
                               train_loss=train_loss,
                               val_loss=val_loss)
 
-            if lr_scheduler is not None and not lr_step_done:
-                lr_scheduler.step()
+            if lr_scheduler and lr_scheduler_step_policy == "epoch":
+                if val_loader and isinstance(lr_scheduler, torch.optim.lr_scheduler.ReduceLROnPlateau):
+                    lr_scheduler.step(val_loss)
+                else:
+                    lr_scheduler.step()
 
             self.writer.flush()
-            if epoch % save_model_after_every_epoch == 0:
-                model_file_name = "model_epoch_{}.pt".format(epoch)
+            print(message)
+            if self.epochs_completed % save_model_after_every_epoch == 0:
+                model_file_name = "model_epoch_{}.pt".format(self.epochs_completed)
                 self.save(model_file_name, save_optimizer_state=True, epoch=self.epochs_completed,
                           train_loss=train_loss, val_loss=val_loss)
 
@@ -359,12 +375,11 @@ class Learner:
                   train_loss=train_loss, val_loss=self.best_val_loss)
 
     def predict(self, loader):
-        predictions, targets = self.__predictor.predict(loader, use_gpu=str(self.device) == "cuda:0")
+        predictions, targets = self.__predictor.predict(loader)
         return predictions, targets
 
     def predict_class(self, loader):
-        predicted_class, probability, targets = self.__predictor.predict_class(loader,
-                                                                               use_gpu=str(self.device) == "cuda:0")
+        predicted_class, probability, targets = self.__predictor.predict_class(loader)
         return predicted_class, probability, targets
 
     def extract_features(self, loader, no_of_features, features_csv_file, iterations=1,
@@ -386,9 +401,9 @@ class Learner:
         with torch.no_grad():
             for iteration in range(iterations):
                 print('Iteration:', iteration + 1)
-                for X, y in tqdm(loader, total=len(loader), desc='Feature Extraction'):
-                    X = X.to(self.device)
-                    feature_set = self.__model(X).cpu().numpy()
+                for x, y in tqdm(loader, total=len(loader), desc='Feature Extraction'):
+
+                    feature_set = self.__predictor.predict_batch(x).cpu().numpy()
 
                     if target_known:
                         y = y.numpy().reshape(-1, 1)
@@ -398,7 +413,8 @@ class Learner:
                     fp.flush()
         fp.close()
 
-    def show_predictions(self, loader, image_inverse_transform=None, samples=9, cols=3, figsize=(10, 10)):
+    def show_predictions(self, loader, image_inverse_transform=None, samples=9, cols=3, figsize=(10, 10),
+                         target_known=True):
 
         self.__predictor.show_predictions(loader, image_inverse_transform=image_inverse_transform,
-                                          samples=samples, cols=cols, figsize=figsize)
+                                          samples=samples, cols=cols, figsize=figsize, target_known=target_known)
